import base64
from typing import TypeVar, Type

T = TypeVar('T')

Separators = (tuple[str] | tuple[str, str] | tuple[str, str, str] | tuple[str, str, str, str]
              | tuple[str, str, str, str, str])
ParseResult = T | list[T | list[T | list[T | list[T | list[T]]]]]


def parse_string(string: str, separators: Separators, target_type: Type[T]) -> ParseResult[T]:
    r"""
    Parse a string (split by separators and convert to the given type).

    The string will be split by the given separators. The amount of separators will determine the dimension of the list.
    For example, if separators = ("\\n", ","), the content will be split by newline and each resulting substring will be
    split by comma, forming a two-dimensional list.
    The substrings will then be converted to the given type (default: str) and returned.
    The function can handle any number of separators/dimensions, but since recursive typing is not widely supported,
    the type hinting is limited to five dimensions.

    :param string: The string to parse.
    :param separators: The separators to split the content by.
    :param target_type: The type to convert the content to.
    :return: The result of the parsing and conversion.
    """
    processed_data: ParseResult = string

    if not separators:
        if target_type != str:
            processed_data = target_type(processed_data)
        return processed_data

    if separators[0] == "":
        processed_data = list(processed_data)
    else:
        processed_data = processed_data.split(separators[0])

    return [parse_string(substr, separators=separators[1:], target_type=target_type) for substr in processed_data]


def parse_file_content(filename: str, separators: Separators, target_type: type[T]) -> ParseResult[T]:
    r"""
    Read file content and parse it (split by separators and convert to the given type).

    The string will be read from the file with the given filename.
    The string will be split by the given separators. The amount of separators will determine the dimension of the list.
    For example, if separators = ("\\n", ","), the content will be split by newline and each resulting substring will be
    split by comma, forming a two-dimensional list.
    The substrings will then be converted to the given type (default: str) and returned.
    The function can handle any number of separators/dimensions, but since recursive typing is not widely supported,
    the type hinting is limited to five dimensions.

    :param filename: The name of the file to read the content from.
    :param separators: The separators to split the content by.
    :param target_type: The type to convert the content to.
    :return: The result of the parsing and conversion.
    """
    if target_type is None:
        target_type = str

    with open(filename) as file:
        file_content = file.read()

    return parse_string(file_content, separators, target_type)


def b64encode(text: str, times_to_encode: int = 1) -> str:
    """
    Encode the given text using base64 encoding.
    :param text: The text to encode.
    :param times_to_encode: The number of times to encode the text.
    :return: The encoded text.
    """
    for _ in range(times_to_encode):
        text = base64.b64encode(text.encode("utf-8")).decode("utf-8")

    return text


def b64decode(text: str, times_to_decode: int = 1) -> str:
    """
    Decode the given text using base64 encoding.
    :param text: The text to decode.
    :param times_to_decode: The number of times to decode the text.
    :return: The decoded text.
    """
    for _ in range(times_to_decode):
        text = base64.b64decode(text).decode("utf-8")

    return text



import linecache
import os
import tracemalloc

BYTES_IN_MEBIBYTE = 1024 ** 2


class CaptureMalloc:
    """
    Context manager to capture memory allocations using ``tracemalloc``.
    
    Upon exiting the context manager, the snapshot is taken and stored in the ``snapshot`` attribute.
    The snapshot is also formatted into a string and stored in the ``snapshot_string`` attribute.
    """

    def __init__(self, max_lines: int = 3):
        """
        Initialize the context manager with the given maximum amount of code lines to include in the snapshot string (sorted by memory usage).
        
        :param max_lines: The maximum amount of code lines to include in the snapshot string. 
        """
        self._max_lines = max_lines
        self.snapshot = None
        self.snapshot_string = ''

    def __enter__(self):
        tracemalloc.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.snapshot = tracemalloc.take_snapshot()
        tracemalloc.stop()
        self.snapshot_string = tm_snapshot_to_string(self.snapshot, limit=self._max_lines)


def tm_snapshot_to_string(snapshot: tracemalloc.Snapshot, key_type='lineno', limit: int = 3) -> str:
    """
    Build a readable string from the given ``tracemalloc`` snapshot.

    :param snapshot: The snapshot to create a string from.
    :param key_type: The key type to group statistics by.
    :param limit: The amount of memory allocations to include in the string.
    :return: The formatted snapshot string.
    """
    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, "<unknown>"),
    ))
    top_stats = snapshot.statistics(key_type)
    report = ["Top %s lines" % limit]

    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        filename = os.sep.join(frame.filename.split(os.sep)[-2:])
        report.append("#%s: %s:%s: %.1f MiB"
                      % (index, filename, frame.lineno, stat.size / BYTES_IN_MEBIBYTE))
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            report.append('    %s' % line)

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        report.append("%s other: %.1f MiB" % (len(other), size / BYTES_IN_MEBIBYTE))
    total = sum(stat.size for stat in top_stats)
    report.append("Total allocated size: %.1f MiB" % (total / BYTES_IN_MEBIBYTE))

    return '\n'.join(report)



"""Utility decorators for Python."""
import inspect
import time
import typing
from functools import wraps


def stopwatch(func):
    """
    Print the runtime of a function.

    It will be printed out like: "It took [time] seconds for [function_name] to run",
    where [time] is the number of seconds (with the precision of at least 5 decimal places)
    it took for the function to run and [function_name] is the name of the function.
    The function's return value will not be affected.

    :param func: The decorated function.
    :return: Inner function.
    """
    recursion_counter = 0
    start = 0

    @wraps(func)
    def wrapper(*args, **kwargs):
        nonlocal recursion_counter, start

        if recursion_counter == 0:
            start = time.perf_counter()

        recursion_counter += 1
        result = func(*args, **kwargs)
        recursion_counter -= 1

        if recursion_counter == 0:
            end = time.perf_counter()
            seconds = end - start
            print(f"It took {seconds} seconds for {func.__name__} to run")

        return result

    return wrapper


def catch(*error_classes):
    """
    Catch the specified exceptions.

    If the function raises one of the specified exceptions, return a tuple of (1, exception_object),
    where exception_object is the caught exception. Otherwise, return a tuple of (0, result),
    where result is the result of the function.

    This decorator is able to handle the following cases:
    1. The decorator is used with no arguments, e.g. @catch. Such usage will catch all exceptions.
    2. The decorator is used with one argument, e.g. @catch(ValueError).
    3. The decorator is used with multiple arguments, e.g. @catch(KeyError, TypeError).
    :param error_classes: The exceptions to catch.
    :return: Inner function.
    """

    def inner(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
                return 0, result
            except error_classes as e:
                return 1, e

        return wrapper

    if len(error_classes) == 1 and error_classes[0].__class__ != type:
        function = error_classes[0]
        error_classes = Exception
        return inner(function)

    return inner


def enforce_types(func):
    """
    Enforce the types of the function's parameters and return value.

    If the function is called with an argument of the wrong type, raise a TypeError with the message:
    "Argument '[argument_name]' must be of type [expected_type], but was [value] of type [actual_type]".
    If the function returns a value of the wrong type, raise a TypeError with the message:
    "Returned value must be of type [expected_type], but was [value] of type [actual_type]".

    If an argument or the return value can be of multiple types, then the [expected_type]
    in the error message will be "[type_1], [type_2], ..., [type_(n-1)] or [type_n]".
    For example if the type annotation for an argument is int | float | str | bool, then the error message will be
    "Argument '[argument_name]' must be of type int, float, str or bool, but was [value] of type [actual_type]".

    If there's no type annotation for a parameter or the return value, then it can be of any type.

    Exceptions, that happen during the execution of the function still occur normally,
    if the argument types are correct.
    :param func: The decorated function.
    :return: Inner function.
    :raises TypeError: If the function is called with an argument of the wrong type or returns a value of the wrong type.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        sig = inspect.signature(func)
        bound = sig.bind(*args, **kwargs)
        err_message_arg = "Argument '{}' must be of type {}, but was {} of type {}"
        err_message_return = "Returned value must be of type {}, but was {} of type {}"

        for name, val in bound.arguments.items():
            if name in func.__annotations__:
                expected_type = func.__annotations__[name]
                verify_type(err_message_arg, expected_type, val, name)

        result = func(*args, **kwargs)

        if (expected_type := func.__annotations__.get("return", 0)) != 0:
            verify_type(err_message_return, expected_type, result)

        return result

    def verify_type(err_message, expected_type, value, parameter_name=None):
        if expected_type is None:
            expected_type = type(None)
        if not isinstance(value, expected_type):
            if typing.get_origin(expected_type) is typing.Union:
                exp_types = tuple(t.__name__ for t in expected_type.__args__)
                expected_type = ', '.join(exp_types[:-1]) + " or " + exp_types[-1]
            else:
                expected_type = expected_type.__name__
            actual_type = type(value).__name__
            if isinstance(value, str):
                value = f"'{value}'"

            if parameter_name is None:
                raise TypeError(err_message.format(expected_type, value, actual_type))
            raise TypeError(err_message.format(parameter_name, expected_type, value, actual_type))

    return wrapper



from ._decorators import stopwatch, catch, enforce_types
from ._dict_to_object import dict_to_object
from ._functions import parse_file_content, parse_string, b64encode, b64decode
from ._memory_profiler import tm_snapshot_to_string, CaptureMalloc



import sys
import types
import typing
from typing import Any, get_type_hints, get_origin, get_args, Collection, Literal, TypeVar

__all__ = ['dict_to_object']

T = TypeVar('T')


def dict_to_object(source_dict: dict[str, Any], target_type: type[T]) -> T:
    """
    Convert a dictionary to an object of the specified type.
    
    This function will create an instance of the specified type and populate its attributes with the values from the
    dictionary. The dictionary keys must match the attribute names of the target type.  
    The target type must have a parameterless constructor (or default values for all parameters).
    
    The following types are supported:
    - Primary types (str, int, float, bool, NoneType)
    - Custom classes (fields have to be defined like in dataclasses), can be nested
    - Collection types (list, set, tuple, dict), can be nested, contain primary types or custom classes
    - Literal types
    - Type aliases, including TypeAliasType (Python 3.12+)
    
    The following types are not supported:
    - Union types
    - Optional types
    
    :param source_dict: The dictionary to convert to an object.
    :param target_type: The type of the object to create.
    :return: 
    """
    constructor = target_type.__init__

    if isinstance(constructor, types.FunctionType):
        code_obj: types.CodeType = constructor.__code__
        arg_count = code_obj.co_argcount + code_obj.co_kwonlyargcount
        defaults_count = len(constructor.__defaults__) + len(constructor.__kwdefaults__)
        if arg_count - defaults_count != 1:
            raise TypeError(
                f"Class {target_type.__qualname__} must have a parameterless constructor (or default values for all parameters).")
    else:
        assert isinstance(constructor, types.WrapperDescriptorType), "Unexpected constructor type."

    target_instance = target_type()

    for attr, sub_type in get_type_hints(target_type).items():
        sub_source = source_dict[attr]
        sub_instance = transform_element(sub_source, sub_type)
        setattr(target_instance, attr, sub_instance)

    return target_instance


def handle_collection(source_collection: Collection[Any], _type: type):
    collection = _type()
    origin = get_origin(_type)
    type_args = get_args(_type)

    if issubclass(origin, list):
        sub_type = type_args[0]

        for sub_source in source_collection:
            sub_instance = transform_element(sub_source, sub_type)
            collection.append(sub_instance)

    elif issubclass(origin, tuple):
        if len(type_args) == 2 and type_args[1] is Ellipsis:
            sub_type = type_args[0]

            for sub_source in source_collection:
                sub_instance = transform_element(sub_source, sub_type)
                collection += (sub_instance,)

        else:
            for sub_type, sub_source in zip(type_args, source_collection):
                sub_instance = transform_element(sub_source, sub_type)
                collection += (sub_instance,)

    elif issubclass(origin, set):
        sub_type = type_args[0]

        for sub_source in source_collection:
            sub_instance = transform_element(sub_source, sub_type)
            collection.add(sub_instance)

    elif issubclass(origin, dict):
        key_type, value_type = type_args
        assert isinstance(source_collection, dict)

        for key_source, value_source in source_collection.items():
            key_instance = transform_element(key_source, key_type)
            value_instance = transform_element(value_source, value_type)

            collection[key_instance] = value_instance

    return collection


def transform_element(sub_source: Collection[Any], sub_type: type):
    if sys.version_info >= (3, 12) and isinstance(sub_type, typing.TypeAliasType):
        sub_type = sub_type.__value__

    sub_origin = get_origin(sub_type) or sub_type

    if sub_origin is types.UnionType:
        raise TypeError("Union types are not supported.")

    if sub_origin is Literal:
        if sub_source not in get_args(sub_type):
            raise ValueError(f"Value {sub_source} is not a valid literal for type {sub_type}.")
        sub_instance = sub_source
    elif is_primary_type(sub_type):
        sub_instance = sub_type(sub_source)
    elif is_collection_type(sub_origin):
        sub_instance = handle_collection(sub_source, sub_type)
    else:
        assert isinstance(sub_source, dict)
        sub_instance = dict_to_object(sub_source, sub_type)

    return sub_instance


def is_collection_type(origin: type):
    return any(issubclass(origin, t) for t in {list, set, tuple, dict})


def is_primary_type(_type: type):
    return any(issubclass(_type, t) for t in {str, int, float, bool, type(None)})
